"""Sports DB league model."""

# pylint: disable=line-too-long
import datetime
from typing import Iterator

import requests_cache
import tqdm

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from ..season_type import SeasonType
from .sportsdb_game_model import create_sportsdb_game_model


class SportsDBLeagueModel(LeagueModel):
    """SportsDB implementation of the league model."""

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        session: requests_cache.CachedSession,
        league_id: str,
        league: League,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        self._league_id = league_id

    @classmethod
    def name(cls) -> str:
        return "sportsdb-league-model"

    def _produce_games(
        self,
        round_str: str,
        week: int,
        league_id: str,
        season_year: str,
        season_type: SeasonType,
        pbar: tqdm.tqdm,
    ) -> Iterator[GameModel]:
        # pylint: disable=line-too-long
        year = int(season_year.split("-")[0])

        def internal_produce_games() -> Iterator[GameModel]:
            response = self.session.get(
                f"https://www.thesportsdb.com/api/v1/json/3/eventsround.php?id={league_id}&r={round_str}&s={season_year}"
            )
            response.raise_for_status()
            games = response.json()
            events = games["events"]
            if events is None:
                return
            for count, game in enumerate(events):
                pbar.update(1)
                game_model = create_sportsdb_game_model(
                    self.session,
                    game,
                    week,
                    count,
                    self.league,
                    year,
                    season_type,
                )
                pbar.set_description(
                    f"SportsDB {season_year} - {season_type} - {game_model.dt}"
                )
                yield game_model

        if year < datetime.datetime.now().year - 1:
            yield from internal_produce_games()
        else:
            with self.session.cache_disabled():
                yield from internal_produce_games()

    @property
    def games(self) -> Iterator[GameModel]:
        with self.session.cache_disabled():
            response = self.session.get(
                f"https://www.thesportsdb.com/api/v1/json/3/search_all_seasons.php?id={self._league_id}"
            )
            response.raise_for_status()
            seasons = response.json()
        with tqdm.tqdm(position=self.position) as pbar:
            for season in seasons["seasons"]:
                season_year = season["strSeason"]
                for season_type in SeasonType:
                    match season_type:
                        case SeasonType.OFFSEASON:
                            yield from self._produce_games(
                                str(0),
                                0,
                                self._league_id,
                                season_year,
                                season_type,
                                pbar,
                            )
                        case SeasonType.PRESEASON:
                            try:
                                yield from self._produce_games(
                                    str(500),
                                    0,
                                    self._league_id,
                                    season_year,
                                    season_type,
                                    pbar,
                                )
                            except ValueError:
                                pass
                        case SeasonType.REGULAR:
                            if self.league != League.NBA:
                                try:
                                    for count, round_str in enumerate(range(1, 125)):
                                        yield from self._produce_games(
                                            str(round_str),
                                            count,
                                            self._league_id,
                                            season_year,
                                            season_type,
                                            pbar,
                                        )
                                except ValueError:
                                    pass
                        case SeasonType.POSTSEASON:
                            for count, round_str in enumerate(
                                [125, 150, 160, 170, 180, 200]
                            ):
                                try:
                                    yield from self._produce_games(
                                        str(round_str),
                                        21 + count,
                                        self._league_id,
                                        season_year,
                                        season_type,
                                        pbar,
                                    )
                                except ValueError:
                                    pass



"""Combined team model."""

# pylint: disable=too-many-locals,too-many-branches,too-many-statements
import re
import unicodedata

from ..coach_model import CoachModel
from ..news_model import NewsModel
from ..odds_model import OddsModel
from ..player_model import PlayerModel
from ..social_model import SocialModel
from ..team_model import TeamModel
from .combined_coach_model import create_combined_coach_model
from .combined_player_model import create_combined_player_model
from .null_check import is_null

REGEX = re.compile("[^a-zA-Z]")


def _normalise_name(name: str) -> str:
    # Handle "Surname, Firstname"
    if "," in name:
        name = " ".join(reversed([x.strip() for x in name.split(",")]))
    return REGEX.sub("", unicodedata.normalize("NFC", name).lower().strip())


def create_combined_team_model(
    team_models: list[TeamModel],
    identifier: str,
    player_identity_map: dict[str, str],
    names: dict[str, str],
    coach_names: dict[str, str],
) -> TeamModel:
    """Create a team model by combining many team models."""
    location = None
    players: dict[str, list[PlayerModel]] = {}
    odds: dict[str, list[OddsModel]] = {}
    news: dict[str, NewsModel] = {}
    social: dict[str, SocialModel] = {}
    coaches: dict[str, list[CoachModel]] = {}
    points = None
    ladder_rank = None
    field_goals = None
    lbw = None
    end_dt = None
    for team_model in team_models:
        team_model_location = team_model.location
        if team_model_location is not None:
            location = team_model_location
        for player_model in team_model.players:
            player_id = player_model.identifier
            player_name_key = _normalise_name(player_model.name)
            if player_model.identifier in player_identity_map:
                player_id = player_identity_map[player_id]
            elif player_name_key in names:
                player_id = names[player_name_key]
            else:
                names[player_name_key] = player_id
            players[player_id] = players.get(player_id, []) + [player_model]
        for odds_model in team_model.odds:
            key = f"{odds_model.bookie.identifier}-{odds_model.odds}"
            odds[key] = odds.get(key, []) + [odds_model]
        team_model_points = team_model.points
        if not is_null(team_model_points):
            points = team_model_points
        team_model_ladder_rank = team_model.ladder_rank
        if not is_null(team_model_ladder_rank):
            ladder_rank = team_model_ladder_rank
        for news_model in team_model.news:
            news_key = "-".join(
                [
                    news_model.title,
                    str(news_model.published),
                    news_model.summary,
                    news_model.source,
                ]
            )
            news[news_key] = news_model
        for social_model in team_model.social:
            social_key = "-".join(
                [social_model.network, social_model.post, str(social_model.published)]
            )
            social[social_key] = social_model
        team_model_field_goals = team_model.field_goals
        if not is_null(team_model_field_goals):
            field_goals = team_model_field_goals
        for coach_model in team_model.coaches:
            coach_id = coach_model.identifier
            coach_name_key = _normalise_name(coach_model.name)
            if coach_name_key in coach_names:
                coach_id = coach_names[coach_name_key]
            else:
                coach_names[coach_name_key] = coach_id
            coaches[coach_id] = coaches.get(coach_id, []) + [coach_model]
        team_model_lbw = team_model.lbw
        if not is_null(team_model_lbw):
            lbw = team_model_lbw
        team_model_end_dt = team_model.end_dt
        if not is_null(team_model_end_dt):
            end_dt = team_model_end_dt

    return TeamModel(
        identifier=identifier,
        name=team_models[0].name,
        location=location,
        players=[  # pyright: ignore
            create_combined_player_model(v, k) for k, v in players.items()
        ],
        odds=[x[0] for x in odds.values()],
        points=points,
        ladder_rank=ladder_rank,
        news=sorted(news.values(), key=lambda x: x.published),
        social=sorted(social.values(), key=lambda x: x.published),
        field_goals=field_goals,
        coaches=[create_combined_coach_model(v, k) for k, v in coaches.items()],
        lbw=lbw,
        end_dt=end_dt,
    )



"""The prototype class for a team."""

# pylint: disable=duplicate-code
import datetime
from typing import Any, Literal

from pydantic import BaseModel, Field

from .coach_model import CoachModel
from .field_type import TYPE_KEY, FieldType
from .news_model import NewsModel
from .odds_model import OddsModel
from .player_model import PlayerModel
from .social_model import SocialModel

TEAM_POINTS_COLUMN: Literal["points"] = "points"
TEAM_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
PLAYER_COLUMN_PREFIX: Literal["players"] = "players"
NAME_COLUMN: Literal["name"] = "name"
FIELD_GOALS_COLUMN: Literal["field_goals"] = "field_goals"
FIELD_GOALS_ATTEMPTED_COLUMN: Literal["field_goals_attempted"] = "field_goals_attempted"
OFFENSIVE_REBOUNDS_COLUMN: Literal["offensive_rebounds"] = "offensive_rebounds"
ASSISTS_COLUMN: Literal["assists"] = "assists"
TURNOVERS_COLUMN: Literal["turnovers"] = "turnovers"
KICKS_COLUMN: Literal["kicks"] = "kicks"
TEAM_ODDS_COLUMN: Literal["odds"] = "odds"
TEAM_MARKS_COLUMN: Literal["marks"] = "marks"
TEAM_HANDBALLS_COLUMN: Literal["handballs"] = "handballs"
TEAM_DISPOSALS_COLUMN: Literal["disposals"] = "disposals"
TEAM_GOALS_COLUMN: Literal["goals"] = "goals"
TEAM_BEHINDS_COLUMN: Literal["behinds"] = "behinds"
TEAM_HIT_OUTS_COLUMN: Literal["hit_outs"] = "hit_outs"
TEAM_TACKLES_COLUMN: Literal["tackles"] = "tackles"
TEAM_REBOUNDS_COLUMN: Literal["rebounds"] = "rebounds"
TEAM_INSIDES_COLUMN: Literal["insides"] = "insides"
TEAM_CLEARANCES_COLUMN: Literal["clearances"] = "clearances"
TEAM_CLANGERS_COLUMN: Literal["clangers"] = "clangers"
TEAM_FREE_KICKS_FOR_COLUMN: Literal["free_kicks_for"] = "free_kicks_for"
TEAM_FREE_KICKS_AGAINST_COLUMN: Literal["free_kicks_against"] = "free_kicks_against"
TEAM_BROWNLOW_VOTES_COLUMN: Literal["brownlow_votes"] = "brownlow_votes"
TEAM_CONTESTED_POSSESSIONS_COLUMN: Literal["contested_possessions"] = (
    "contested_possessions"
)
TEAM_UNCONTESTED_POSSESSIONS_COLUMN: Literal["uncontested_possessions"] = (
    "uncontested_possessions"
)
TEAM_CONTESTED_MARKS_COLUMN: Literal["contested_marks"] = "contested_marks"
TEAM_MARKS_INSIDE_COLUMN: Literal["marks_inside"] = "marks_inside"
TEAM_ONE_PERCENTERS_COLUMN: Literal["one_percenters"] = "one_percenters"
TEAM_BOUNCES_COLUMN: Literal["bounces"] = "bounces"
TEAM_GOAL_ASSISTS_COLUMN: Literal["goal_assists"] = "goal_assists"
TEAM_NEWS_COLUMN: Literal["news"] = "news"
TEAM_COACHES_COLUMN: Literal["coaches"] = "coaches"
TEAM_LENGTH_BEHIND_WINNER_COLUMN: Literal["lbw"] = "lbw"
TEAM_END_DT_COLUMN: Literal["end_dt"] = "end_dt"


def _calculate_kicks(data: dict[str, Any]) -> int | None:
    kicks = 0
    found_kicks = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_kicks = player.kicks
        if player_kicks is None:
            continue
        found_kicks = True
        kicks += player_kicks
    if not found_kicks:
        return None
    return kicks


def _calculate_field_goals(data: dict[str, Any]) -> int | None:
    field_goals = 0
    found_field_goals = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_field_goals = player.field_goals
        if player_field_goals is None:
            continue
        found_field_goals = True
        field_goals += player_field_goals
    if not found_field_goals:
        return None
    return field_goals


def _calculate_field_goals_attempted(data: dict[str, Any]) -> int | None:
    field_goals_attempted = 0
    found_field_goals_attempted = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_field_goals_attempted = player.field_goals_attempted
        if player_field_goals_attempted is None:
            continue
        found_field_goals_attempted = True
        field_goals_attempted += player_field_goals_attempted
    if not found_field_goals_attempted:
        return None
    return field_goals_attempted


def _calculate_offensive_rebounds(data: dict[str, Any]) -> int | None:
    offensive_rebounds = 0
    found_offensive_rebounds = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_offensive_rebounds = player.offensive_rebounds
        if player_offensive_rebounds is None:
            continue
        found_offensive_rebounds = True
        offensive_rebounds += player_offensive_rebounds
    if not found_offensive_rebounds:
        return None
    return offensive_rebounds


def _calculate_assists(data: dict[str, Any]) -> int | None:
    assists = 0
    found_assists = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_assists = player.assists
        if player_assists is None:
            continue
        found_assists = True
        assists += player_assists
    if not found_assists:
        return None
    return assists


def _calculate_turnovers(data: dict[str, Any]) -> int | None:
    turnovers = 0
    found_turnovers = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_turnovers = player.turnovers
        if player_turnovers is None:
            continue
        found_turnovers = True
        turnovers += player_turnovers
    if not found_turnovers:
        return None
    return turnovers


def _calculate_marks(data: dict[str, Any]) -> int | None:
    marks = 0
    found_marks = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_marks = player.marks
        if player_marks is None:
            continue
        found_marks = True
        marks += player_marks
    if not found_marks:
        return None
    return marks


def _calculate_handballs(data: dict[str, Any]) -> int | None:
    handballs = 0
    found_handballs = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_handballs = player.handballs
        if player_handballs is None:
            continue
        found_handballs = True
        handballs += player_handballs
    if not found_handballs:
        return None
    return handballs


def _calculate_disposals(data: dict[str, Any]) -> int | None:
    disposals = 0
    found_disposals = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_disposals = player.disposals
        if player_disposals is None:
            continue
        found_disposals = True
        disposals += player_disposals
    if not found_disposals:
        return None
    return disposals


def _calculate_goals(data: dict[str, Any]) -> int | None:
    goals = 0
    found_goals = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_goals = player.goals
        if player_goals is None:
            continue
        found_goals = True
        goals += player_goals
    if not found_goals:
        return None
    return goals


def _calculate_behinds(data: dict[str, Any]) -> int | None:
    behinds = 0
    found_behinds = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_behinds = player.behinds
        if player_behinds is None:
            continue
        found_behinds = True
        behinds += player_behinds
    if not found_behinds:
        return None
    return behinds


def _calculate_hit_outs(data: dict[str, Any]) -> int | None:
    hit_outs = 0
    found_hit_outs = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_hit_outs = player.hit_outs
        if player_hit_outs is None:
            continue
        found_hit_outs = True
        hit_outs += player_hit_outs
    if not found_hit_outs:
        return None
    return hit_outs


def _calculate_tackles(data: dict[str, Any]) -> int | None:
    tackles = 0
    found_tackles = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_tackles = player.tackles
        if player_tackles is None:
            continue
        found_tackles = True
        tackles += player_tackles
    if not found_tackles:
        return None
    return tackles


def _calculate_rebounds(data: dict[str, Any]) -> int | None:
    rebounds = 0
    found_rebounds = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_rebounds = player.rebounds
        if player_rebounds is None:
            continue
        found_rebounds = True
        rebounds += player_rebounds
    if not found_rebounds:
        return None
    return rebounds


def _calculate_insides(data: dict[str, Any]) -> int | None:
    insides = 0
    found_insides = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_insides = player.insides
        if player_insides is None:
            continue
        found_insides = True
        insides += player_insides
    if not found_insides:
        return None
    return insides


def _calculate_clearances(data: dict[str, Any]) -> int | None:
    clearances = 0
    found_clearances = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_clearances = player.clearances
        if player_clearances is None:
            continue
        found_clearances = True
        clearances += player_clearances
    if not found_clearances:
        return None
    return clearances


def _calculate_clangers(data: dict[str, Any]) -> int | None:
    clangers = 0
    found_clangers = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_clangers = player.clangers
        if player_clangers is None:
            continue
        found_clangers = True
        clangers += player_clangers
    if not found_clangers:
        return None
    return clangers


def _calculate_free_kicks_for(data: dict[str, Any]) -> int | None:
    free_kicks_for = 0
    found_free_kicks_for = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_free_kicks_for = player.free_kicks_for
        if player_free_kicks_for is None:
            continue
        found_free_kicks_for = True
        free_kicks_for += player_free_kicks_for
    if not found_free_kicks_for:
        return None
    return free_kicks_for


def _calculate_free_kicks_against(data: dict[str, Any]) -> int | None:
    free_kicks_against = 0
    found_free_kicks_against = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_free_kicks_against = player.free_kicks_against
        if player_free_kicks_against is None:
            continue
        found_free_kicks_against = True
        free_kicks_against += player_free_kicks_against
    if not found_free_kicks_against:
        return None
    return free_kicks_against


def _calculate_brownlow_votes(data: dict[str, Any]) -> int | None:
    brownlow_votes = 0
    found_brownlow_votes = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_brownlow_votes = player.brownlow_votes
        if player_brownlow_votes is None:
            continue
        found_brownlow_votes = True
        brownlow_votes += player_brownlow_votes
    if not found_brownlow_votes:
        return None
    return brownlow_votes


def _calculate_contested_possessions(data: dict[str, Any]) -> int | None:
    contested_possessions = 0
    found_contested_possessions = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_contested_possessions = player.contested_possessions
        if player_contested_possessions is None:
            continue
        found_contested_possessions = True
        contested_possessions += player_contested_possessions
    if not found_contested_possessions:
        return None
    return contested_possessions


def _calculate_uncontested_possessions(data: dict[str, Any]) -> int | None:
    uncontested_possessions = 0
    found_uncontested_possessions = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_uncontested_possessions = player.uncontested_possessions
        if player_uncontested_possessions is None:
            continue
        found_uncontested_possessions = True
        uncontested_possessions += player_uncontested_possessions
    if not found_uncontested_possessions:
        return None
    return uncontested_possessions


def _calculate_contested_marks(data: dict[str, Any]) -> int | None:
    contested_marks = 0
    found_contested_marks = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_contested_marks = player.contested_marks
        if player_contested_marks is None:
            continue
        found_contested_marks = True
        contested_marks += player_contested_marks
    if not found_contested_marks:
        return None
    return contested_marks


def _calculate_marks_inside(data: dict[str, Any]) -> int | None:
    marks_inside = 0
    found_marks_inside = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_marks_inside = player.marks_inside
        if player_marks_inside is None:
            continue
        found_marks_inside = True
        marks_inside += player_marks_inside
    if not found_marks_inside:
        return None
    return marks_inside


def _calculate_one_percenters(data: dict[str, Any]) -> int | None:
    one_percenters = 0
    found_one_percenters = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_one_percenters = player.one_percenters
        if player_one_percenters is None:
            continue
        found_one_percenters = True
        one_percenters += player_one_percenters
    if not found_one_percenters:
        return None
    return one_percenters


def _calculate_bounces(data: dict[str, Any]) -> int | None:
    bounces = 0
    found_bounces = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_bounces = player.bounces
        if player_bounces is None:
            continue
        found_bounces = True
        bounces += player_bounces
    if not found_bounces:
        return None
    return bounces


def _calculate_goal_assists(data: dict[str, Any]) -> int | None:
    goal_assists = 0
    found_goal_assists = False
    for player in data.get(PLAYER_COLUMN_PREFIX, []):
        player_goal_assists = player.goal_assists
        if player_goal_assists is None:
            continue
        found_goal_assists = True
        goal_assists += player_goal_assists
    if not found_goal_assists:
        return None
    return goal_assists


class TeamModel(BaseModel):
    """The serialisable team class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=TEAM_IDENTIFIER_COLUMN,
    )
    name: str = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.TEXT}, alias=NAME_COLUMN
    )
    location: str | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL}
    )
    players: list[PlayerModel] = Field(..., alias=PLAYER_COLUMN_PREFIX)
    odds: list[OddsModel] = Field(..., alias=TEAM_ODDS_COLUMN)
    points: float | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.POINTS}, alias=TEAM_POINTS_COLUMN
    )
    ladder_rank: int | None
    kicks: int | None = Field(
        default_factory=_calculate_kicks,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=KICKS_COLUMN,
    )
    news: list[NewsModel] = Field(..., alias=TEAM_NEWS_COLUMN)
    social: list[SocialModel]
    field_goals: int | None = Field(
        default_factory=_calculate_field_goals,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=FIELD_GOALS_COLUMN,
    )
    field_goals_attempted: int | None = Field(
        default_factory=_calculate_field_goals_attempted,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=FIELD_GOALS_ATTEMPTED_COLUMN,
    )
    offensive_rebounds: int | None = Field(
        default_factory=_calculate_offensive_rebounds,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=OFFENSIVE_REBOUNDS_COLUMN,
    )
    assists: int | None = Field(
        default_factory=_calculate_assists,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=ASSISTS_COLUMN,
    )
    turnovers: int | None = Field(
        default_factory=_calculate_turnovers,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TURNOVERS_COLUMN,
    )
    marks: int | None = Field(
        default_factory=_calculate_marks,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_MARKS_COLUMN,
    )
    handballs: int | None = Field(
        default_factory=_calculate_handballs,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_HANDBALLS_COLUMN,
    )
    disposals: int | None = Field(
        default_factory=_calculate_disposals,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_DISPOSALS_COLUMN,
    )
    goals: int | None = Field(
        default_factory=_calculate_goals,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_GOALS_COLUMN,
    )
    behinds: int | None = Field(
        default_factory=_calculate_behinds,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_BEHINDS_COLUMN,
    )
    hit_outs: int | None = Field(
        default_factory=_calculate_hit_outs,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_HIT_OUTS_COLUMN,
    )
    tackles: int | None = Field(
        default_factory=_calculate_tackles,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_TACKLES_COLUMN,
    )
    rebounds: int | None = Field(
        default_factory=_calculate_rebounds,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_REBOUNDS_COLUMN,
    )
    insides: int | None = Field(
        default_factory=_calculate_insides,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_INSIDES_COLUMN,
    )
    clearances: int | None = Field(
        default_factory=_calculate_clearances,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_CLEARANCES_COLUMN,
    )
    clangers: int | None = Field(
        default_factory=_calculate_clangers,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_CLANGERS_COLUMN,
    )
    free_kicks_for: int | None = Field(
        default_factory=_calculate_free_kicks_for,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_FREE_KICKS_FOR_COLUMN,
    )
    free_kicks_against: int | None = Field(
        default_factory=_calculate_free_kicks_against,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_FREE_KICKS_AGAINST_COLUMN,
    )
    brownlow_votes: int | None = Field(
        default_factory=_calculate_brownlow_votes,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_BROWNLOW_VOTES_COLUMN,
    )
    contested_possessions: int | None = Field(
        default_factory=_calculate_contested_possessions,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_CONTESTED_POSSESSIONS_COLUMN,
    )
    uncontested_possessions: int | None = Field(
        default_factory=_calculate_uncontested_possessions,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_UNCONTESTED_POSSESSIONS_COLUMN,
    )
    contested_marks: int | None = Field(
        default_factory=_calculate_contested_marks,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_CONTESTED_MARKS_COLUMN,
    )
    marks_inside: int | None = Field(
        default_factory=_calculate_marks_inside,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_MARKS_INSIDE_COLUMN,
    )
    one_percenters: int | None = Field(
        default_factory=_calculate_one_percenters,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_ONE_PERCENTERS_COLUMN,
    )
    bounces: int | None = Field(
        default_factory=_calculate_bounces,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_BOUNCES_COLUMN,
    )
    goal_assists: int | None = Field(
        default_factory=_calculate_goal_assists,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_GOAL_ASSISTS_COLUMN,
    )
    coaches: list[CoachModel] = Field(..., alias=TEAM_COACHES_COLUMN)
    lbw: float | None = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD},
        alias=TEAM_LENGTH_BEHIND_WINNER_COLUMN,
    )
    end_dt: datetime.datetime | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.LOOKAHEAD}, alias=TEAM_END_DT_COLUMN
    )



"""HKJC HKJC team model."""

# pylint: disable=duplicate-code,too-many-arguments,too-many-locals
import datetime

import pytest_is_running
import requests_cache

from ....cache import MEMORY
from ...odds_model import OddsModel
from ...team_model import TeamModel
from ..position import Position
from .hkjc_hkjc_coach_model import create_hkjc_hkjc_coach_model
from .hkjc_hkjc_player_model import create_hkjc_hkjc_player_model


def _create_hkjc_hkjc_team_model(
    session: requests_cache.CachedSession,
    horse_url: str,
    jockey_url: str | None,
    trainer_url: str | None,
    points: float,
    jersey: str,
    handicap_weight: float | None,
    horse_weight: float | None,
    starting_position: Position | None,
    lbw: float | None,
    end_dt: datetime.datetime | None,
    odds: list[OddsModel],
) -> TeamModel:
    players = []
    horse_player = create_hkjc_hkjc_player_model(
        session=session,
        url=horse_url,
        jersey=jersey,
        handicap_weight=handicap_weight,
        starting_position=starting_position,
        weight=horse_weight,
    )
    if horse_player is not None:
        players.append(horse_player)
    if jockey_url is not None:
        jockey_player = create_hkjc_hkjc_player_model(
            session=session,
            url=jockey_url,
            jersey=jersey,
            handicap_weight=None,
            starting_position=starting_position,
            weight=None,
        )
        if jockey_player is not None:
            players.append(jockey_player)
    coaches = (
        [create_hkjc_hkjc_coach_model(session=session, url=trainer_url)]
        if trainer_url is not None
        else []
    )
    name = " - ".join([x.name for x in players])
    return TeamModel(
        identifier=name,
        name=name,
        location=None,
        players=players,
        odds=odds,
        points=points,
        ladder_rank=None,
        news=[],
        social=[],
        field_goals=None,
        coaches=coaches,
        lbw=lbw,
        end_dt=end_dt,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_hkjc_hkjc_team_model(
    session: requests_cache.CachedSession,
    horse_url: str,
    jockey_url: str | None,
    trainer_url: str | None,
    points: float,
    jersey: str,
    handicap_weight: float | None,
    horse_weight: float | None,
    starting_position: Position | None,
    lbw: float | None,
    end_dt: datetime.datetime | None,
    odds: list[OddsModel],
) -> TeamModel:
    return _create_hkjc_hkjc_team_model(
        session=session,
        horse_url=horse_url,
        jockey_url=jockey_url,
        trainer_url=trainer_url,
        points=points,
        jersey=jersey,
        handicap_weight=handicap_weight,
        horse_weight=horse_weight,
        starting_position=starting_position,
        lbw=lbw,
        end_dt=end_dt,
        odds=odds,
    )


def create_hkjc_hkjc_team_model(
    session: requests_cache.CachedSession,
    horse_url: str,
    jockey_url: str | None,
    trainer_url: str | None,
    points: float,
    jersey: str,
    handicap_weight: float | None,
    horse_weight: float | None,
    starting_position: Position | None,
    lbw: float | None,
    end_dt: datetime.datetime | None,
    odds: list[OddsModel],
) -> TeamModel:
    """Create team model from HKJC."""
    if not pytest_is_running.is_running():
        return _cached_create_hkjc_hkjc_team_model(
            session=session,
            horse_url=horse_url,
            jockey_url=jockey_url,
            trainer_url=trainer_url,
            points=points,
            jersey=jersey,
            handicap_weight=handicap_weight,
            horse_weight=horse_weight,
            starting_position=starting_position,
            lbw=lbw,
            end_dt=end_dt,
            odds=odds,
        )
    return _create_hkjc_hkjc_team_model(
        session=session,
        horse_url=horse_url,
        jockey_url=jockey_url,
        trainer_url=trainer_url,
        points=points,
        jersey=jersey,
        handicap_weight=handicap_weight,
        horse_weight=horse_weight,
        starting_position=starting_position,
        lbw=lbw,
        end_dt=end_dt,
        odds=odds,
    )



"""ESPN league model."""

# pylint: disable=too-many-locals,too-many-arguments
from typing import Any, Iterator

import requests_cache
import tqdm

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from ..season_type import SeasonType
from .espn_game_model import create_espn_game_model


def _season_type_from_name(name: str) -> SeasonType:
    if name == "Regular Season":
        return SeasonType.REGULAR
    if name == "Preseason":
        return SeasonType.PRESEASON
    if name == "Postseason":
        return SeasonType.POSTSEASON
    if name == "Off Season":
        return SeasonType.OFFSEASON
    raise ValueError(f"Unrecognised season name: {name}")


class ESPNLeagueModel(LeagueModel):
    """ESPN implementation of the league model."""

    def __init__(
        self,
        start_url: str,
        league: League,
        session: requests_cache.CachedSession,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        self._start_url = start_url

    @classmethod
    def name(cls) -> str:
        return "espn-league-model"

    @classmethod
    def position_validator(cls) -> dict[str, str]:
        """A dictionary that contains the mapping from positions to standard positions."""
        raise NotImplementedError(
            "position_validator is not implemented by parent class"
        )

    def _produce_games(
        self,
        week: dict[str, Any],
        week_count: int,
        season_type_json: dict[str, Any],
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        def produce_game(
            event_item: dict[str, Any],
            game_number: int,
        ) -> GameModel:
            if cache_disabled:
                with self.session.cache_disabled():
                    event_response = self.session.get(event_item["$ref"])
            else:
                event_response = self.session.get(event_item["$ref"])
            event_response.raise_for_status()
            event = event_response.json()
            game_model = create_espn_game_model(
                event,
                week_count,
                game_number,
                self.session,
                self.league,
                season_type_json["year"],
                _season_type_from_name(season_type_json["name"]),
                self.position_validator(),
            )
            pbar.update(1)
            pbar.set_description(
                f"ESPN {game_model.year} - {game_model.season_type} - {game_model.dt}"
            )
            return game_model

        events_page = 1
        events_count = 0
        while True:
            if "events" not in week:
                break
            events_response = self.session.get(
                week["events"]["$ref"] + f"&page={events_page}"
            )
            events_response.raise_for_status()
            events = events_response.json()
            for event_item in events["items"]:
                yield produce_game(event_item, events_count)
                events_count += 1
            if events_page >= events["pageCount"]:
                break
            events_page += 1
        qbr_page = 1
        qbr_count = 0
        while True:
            if "qbr" not in week:
                break
            qbr_response = self.session.get(week["qbr"]["$ref"] + f"&page={qbr_page}")
            qbr = qbr_response.json()
            for qbr_item in qbr["items"]:
                yield produce_game(qbr_item["event"], qbr_count)
                qbr_count += 1
            if qbr_page >= qbr["pageCount"]:
                break
            qbr_page += 1

    def _produce_week_games(
        self,
        season_type_json: dict[str, Any],
        page: int,
        pbar: tqdm.tqdm,
        cache_disabled: bool,
    ) -> Iterator[GameModel]:
        game_page = 1
        week_count = 0
        while True:
            if cache_disabled:
                with self.session.cache_disabled():
                    weeks_response = self.session.get(
                        season_type_json["weeks"]["$ref"] + f"&page={page}"
                    )
            else:
                weeks_response = self.session.get(
                    season_type_json["weeks"]["$ref"] + f"&page={page}"
                )
            weeks_response.raise_for_status()
            weeks = weeks_response.json()
            for item in weeks["items"]:
                if cache_disabled:
                    with self.session.cache_disabled():
                        week_response = self.session.get(item["$ref"])
                else:
                    week_response = self.session.get(item["$ref"])
                week_response.raise_for_status()
                week = week_response.json()
                yield from self._produce_games(
                    week, week_count, season_type_json, pbar, cache_disabled
                )
                week_count += 1
            if game_page >= weeks["pageCount"]:
                break
            game_page += 1

    @property
    def games(self) -> Iterator[GameModel]:
        page = 1
        first = True
        with tqdm.tqdm(position=self.position) as pbar:
            while True:
                if page == 1:
                    with self.session.cache_disabled():
                        response = self.session.get(self._start_url + f"&page={page}")
                else:
                    response = self.session.get(self._start_url + f"&page={page}")
                response.raise_for_status()
                seasons = response.json()
                for item in seasons.get("items", []):
                    season_response = self.session.get(item["$ref"])
                    season_response.raise_for_status()
                    season_json = season_response.json()

                    for season_item in season_json["types"]["items"]:
                        season_type_response = self.session.get(season_item["$ref"])
                        season_type_response.raise_for_status()
                        season_type_json = season_type_response.json()

                        yield from self._produce_week_games(
                            season_type_json, page, pbar, first
                        )
                    first = False

                if page >= seasons.get("pageCount", 0):
                    break
                page += 1



"""OddsPortal bookie model."""

from ...cache import MEMORY
from ..bookie_model import BookieModel


@MEMORY.cache
def create_oddsportal_bookie_model(bookie_name: str, bookie_id: str) -> BookieModel:
    """Create bookie model from odds portal."""
    return BookieModel(identifier=bookie_id, name=bookie_name)



"""A session that rotates proxies."""

# pylint: disable=W0223,too-many-boolean-expressions,protected-access,unused-argument,too-many-arguments
import datetime
import logging
import os
import random
import sqlite3
from io import BytesIO
from typing import Any, MutableMapping, Optional

import numpy as np
import requests
import requests_cache
import urllib3
import wayback  # type: ignore
from func_timeout import FunctionTimedOut, func_set_timeout  # type: ignore
from random_user_agent.params import OperatingSystem  # type: ignore
from random_user_agent.params import SoftwareName
from random_user_agent.user_agent import UserAgent  # type: ignore
from requests import PreparedRequest
from requests.cookies import RequestsCookieJar
from requests.models import Request, Response
from requests_cache import AnyResponse, ExpirationTime
from tenacity import (after_log, before_log, retry, retry_if_exception_type,
                      stop_after_attempt, wait_random_exponential)
from urllib3.response import HTTPResponse

from .session import DEFAULT_TIMEOUT

FAST_FAIL_DOMAINS = [
    "https://news.google.com/",
    "https://historical-forecast-api.open-meteo.com/",
]
X_NO_WAYBACK = "x-no-wayback"


def _is_fast_fail_url(url: str | None) -> bool:
    if url is None:
        return False
    for fast_fail_domain in FAST_FAIL_DOMAINS:
        if url.startswith(fast_fail_domain):
            return True
    return False


class ProxySession(requests_cache.CachedSession):
    """A requests session that can rotate between different proxies."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._proxies = []
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self._last_fetched = None
        software_names = [SoftwareName.CHROME.value]
        operating_systems = [OperatingSystem.WINDOWS.value, OperatingSystem.LINUX.value]
        self._user_agent_rotator = UserAgent(
            software_names=software_names, operating_systems=operating_systems
        )
        self._wayback_client = wayback.WaybackClient()

    def _suggest_proxy(self) -> str:
        proxies = self._proxies
        if (
            not proxies
            or self._last_fetched is None
            or (
                self._last_fetched
                < datetime.datetime.now() - datetime.timedelta(minutes=30)
            )
        ):
            proxies = [""]  # This indicates that no proxy is used
            proxies_txt = os.environ.get("PROXIES")
            if proxies_txt is not None:
                proxies.extend(proxies_txt.split(","))
            random.shuffle(proxies)
            self._proxies = proxies
            self._last_fetched = datetime.datetime.now()

        proxy = random.choices(
            population=proxies,
            weights=np.linspace(1.0, 0.0, len(proxies)).tolist(),
            k=1,
        )
        return proxy[0]

    def _wayback_machine_request(
        self, method, url, **kwargs
    ) -> requests.Response | None:
        if method.upper() != "GET":
            return None
        try:
            for record in self._wayback_client.search(url, fast_latest=True):
                if record.timestamp.replace(
                    tzinfo=None
                ) < datetime.datetime.now().replace(tzinfo=None) - datetime.timedelta(
                    days=350 * 10
                ):
                    continue
                with self._wayback_client.get_memento(record) as memento:
                    cookies = RequestsCookieJar()
                    response = Response()
                    response.status_code = memento.status_code
                    response._content = memento.content
                    response.url = url
                    response.headers = memento.headers
                    response.cookies = cookies
                    response.raw = HTTPResponse(
                        body=BytesIO(memento.content),
                        status=memento.status_code,
                        headers=memento.headers,
                        preload_content=False,
                    )

                    request = Request(
                        method=method,
                        url=url,
                        headers=kwargs.get("headers"),
                        cookies=cookies,
                    )
                    prepared_request = request.prepare()
                    prepared_request.prepare_cookies(cookies)
                    response.request = prepared_request

                    return response
        except (
            wayback.exceptions.MementoPlaybackError,  # pyright: ignore
            requests.exceptions.ChunkedEncodingError,
            requests.exceptions.ContentDecodingError,
        ):  # pyright: ignore
            pass
        return None

    @func_set_timeout(DEFAULT_TIMEOUT)
    def _perform_timeout_send(
        self, request: requests.PreparedRequest, no_wayback: bool, **kwargs
    ) -> Any:
        key = self.cache.create_key(request)

        if not self.settings.disabled:
            # Check the cache
            cached_response = self.cache.get_response(key)
            if cached_response:
                return cached_response

            logging.info(
                "Request for %s not cached (no-wayback: %s).",
                request.url,
                str(no_wayback),
            )

            # Otherwise check the wayback machine
            if not _is_fast_fail_url(request.url) and not no_wayback:
                response = self._wayback_machine_request(
                    request.method, request.url, headers=request.headers
                )
                if response is not None and response.ok:
                    logging.info(
                        "Found wayback machine memento for URL: %s", request.url
                    )
                    self.cache.save_response(response=response, cache_key=key)
                    return response
        else:
            logging.info("Request for %s caching disabled.", request.url)

        response = super().send(request, **kwargs)
        if not _is_fast_fail_url(response.url):
            response.raise_for_status()
        return response

    @retry(
        stop=stop_after_attempt(64),
        after=after_log(logging.getLogger(__name__), logging.DEBUG),
        before=before_log(logging.getLogger(__name__), logging.DEBUG),
        wait=wait_random_exponential(multiplier=1, max=240),
        retry=retry_if_exception_type(FunctionTimedOut)
        | retry_if_exception_type(requests.exceptions.ProxyError)
        | retry_if_exception_type(requests.exceptions.ConnectionError)
        | retry_if_exception_type(requests.exceptions.ChunkedEncodingError)
        | retry_if_exception_type(ValueError)
        | retry_if_exception_type(requests.exceptions.HTTPError)
        | retry_if_exception_type(requests.exceptions.ReadTimeout)
        | retry_if_exception_type(sqlite3.OperationalError),
        reraise=True,
    )
    def _perform_retry_send(
        self, request: requests.PreparedRequest, no_wayback: bool, **kwargs
    ) -> Any:
        return self._perform_timeout_send(request, no_wayback, **kwargs)

    def send(
        self,
        request: PreparedRequest,
        expire_after: ExpirationTime = None,  # pyright: ignore
        only_if_cached: bool = False,
        refresh: bool = False,
        force_refresh: bool = False,
        **kwargs,
    ) -> AnyResponse:
        if _is_fast_fail_url(request.url):
            response = super().send(
                request,
                expire_after=expire_after,
                only_if_cached=only_if_cached,
                refresh=refresh,
                force_refresh=force_refresh,
                **kwargs,
            )
            self.cache.save_response(response)
            return response

        no_wayback = request.headers.get(X_NO_WAYBACK) == "1"
        if request.headers.get(X_NO_WAYBACK) is not None:
            del request.headers[X_NO_WAYBACK]
        return self._perform_retry_send(
            request,
            no_wayback,
            expire_after=expire_after,
            only_if_cached=only_if_cached,
            refresh=refresh,
            force_refresh=force_refresh,
            **kwargs,
        )

    def request(  # type: ignore
        self,
        method: str,
        url: str,
        *args,
        headers: Optional[MutableMapping[str, str]] = None,
        expire_after: ExpirationTime = None,  # pyright: ignore
        only_if_cached: bool = False,
        refresh: bool = False,
        force_refresh: bool = False,
        **kwargs,
    ) -> AnyResponse:
        if "timeout" not in kwargs:
            if _is_fast_fail_url(url):
                kwargs["timeout"] = 5.0
            else:
                kwargs["timeout"] = DEFAULT_TIMEOUT
        if headers is None:
            headers = {}
        if "User-Agent" not in headers:
            headers["User-Agent"] = (
                self._user_agent_rotator.get_random_user_agent().strip()
            )
        proxy = self._suggest_proxy()
        if proxy:
            logging.debug("Using proxy: %s", proxy)
            kwargs.setdefault(
                "proxies",
                {
                    "http": proxy,
                    "https": proxy,
                },
            )
        return super().request(
            method,
            url,
            *args,
            headers=headers,
            expire_after=expire_after,
            only_if_cached=only_if_cached,
            refresh=refresh,
            force_refresh=force_refresh,
            **kwargs,
        )


def create_proxy_session() -> ProxySession:
    """Creates a standard proxy session."""
    return ProxySession(
        "sportsball",
        expire_after=requests_cache.NEVER_EXPIRE,
        allowable_methods=("GET", "HEAD", "POST"),
    )



"""A function for checking whether an object is null."""

from typing import Any

import numpy as np


def is_null(obj: Any) -> bool:
    """Whether the object is a null type object."""
    if obj is None:
        return True
    try:
        if np.isnan(obj):
            return True
    except TypeError:
        pass
    try:
        if np.isnat(obj):
            return True
    except TypeError:
        pass
    return False



"""Aussportsbetting league model."""

# pylint: disable=line-too-long
from io import BytesIO
from typing import Any, Iterator

import requests_cache
import tqdm
from dateutil.parser import parse
from openpyxl import load_workbook

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from .aussportsbetting_game_model import create_aussportsbetting_game_model


class AusSportsBettingLeagueModel(LeagueModel):
    """AusSportsBetting implementation of the league model."""

    def __init__(
        self,
        league: League,
        session: requests_cache.CachedSession,
        position: int | None = None,
    ) -> None:
        super().__init__(league, session, position=position)
        match league:
            case League.AFL:
                self._spreadsheet_url = (
                    "https://www.aussportsbetting.com/historical_data/afl.xlsx"
                )
            case League.NFL:
                self._spreadsheet_url = (
                    "https://www.aussportsbetting.com/historical_data/nfl.xlsx"
                )
            case _:
                raise ValueError(
                    f"League {league} not supported by aus sports betting."
                )

    @classmethod
    def name(cls) -> str:
        return "aussportsbetting-league-model"

    def _row_to_game(self, row: Any) -> GameModel | None:
        current_cell_idx = 0
        date_cell = str(row[0].value)
        if date_cell in {"Date", "None"}:
            return None
        current_cell_idx += 1
        if self.league == League.AFL:
            time_cell = str(row[current_cell_idx].value)
            dt = parse(" ".join([date_cell, time_cell]))
            current_cell_idx += 1
        else:
            dt = parse(date_cell)
        home_team = str(row[current_cell_idx].value).strip()
        current_cell_idx += 1
        away_team = str(row[current_cell_idx].value).strip()
        current_cell_idx += 1
        venue = None
        if self.league == League.AFL:
            venue = str(row[current_cell_idx].value).strip()
            current_cell_idx += 1
        home_points = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        away_points = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1

        if self.league == League.NFL:
            current_cell_idx += 1
        play_off = row[current_cell_idx].value == "Y"

        if self.league == League.NFL:
            current_cell_idx += 2
        else:
            current_cell_idx += 4
        home_odds = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        if self.league == League.NFL:
            current_cell_idx += 3
        away_odds = float(row[current_cell_idx].value)  # type: ignore
        current_cell_idx += 1
        return create_aussportsbetting_game_model(
            dt,
            home_team,
            away_team,
            venue,
            self.session,
            home_points,
            away_points,
            home_odds,
            away_odds,
            self.league,
            play_off,
        )

    @property
    def games(self) -> Iterator[GameModel]:
        with self.session.cache_disabled():
            response = self.session.get(self._spreadsheet_url)
        response.raise_for_status()
        workbook = load_workbook(filename=BytesIO(response.content))
        ws = workbook.active
        if ws is None:
            raise ValueError("ws is null.")
        with tqdm.tqdm(position=self.position) as pbar:
            for row in ws.iter_rows():
                game_model = self._row_to_game(row)
                if game_model is not None:
                    pbar.update(1)
                    pbar.set_description(f"AusSportsBetting - {game_model.dt}")
                    yield game_model



"""The enumeration of the different supported positions."""

from enum import StrEnum


class Position(StrEnum):
    """An enumeration over the different positions."""

    LINE_BACKER = "LB"
    DEFENSIVE_TACKLE = "DT"



"""AFL AFL league model."""

# pylint: disable=too-many-statements,protected-access,too-many-locals,bare-except,too-many-branches,duplicate-code
import datetime
import logging
import re
import urllib.parse
from typing import Iterator

import requests_cache
from bs4 import BeautifulSoup, Tag
from dateutil.parser import parse
from playwright.sync_api import Playwright, sync_playwright

from ....playwright import ensure_install
from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from ..position import Position, position_from_str
from .afl_afl_game_model import create_afl_afl_game_model, parse_players_v1


def _parse_v1(
    soup: BeautifulSoup,
    session: requests_cache.CachedSession,
    ladder: list[str],
    html_url: str,
    playwright: Playwright,
) -> Iterator[GameModel]:
    for div in soup.find_all("div", {"class": re.compile(".*js-match-list-item.*")}):
        team_names = []
        for span in div.find_all(
            "span", {"class": re.compile(".*team-lineups__team-name.*")}
        ):
            team_names.append(span.get_text().strip())
        teams_players = parse_players_v1(div)
        dt = None
        for time in div.find_all(
            "time", {"class": re.compile(".*match-list-alt__header-time.*")}
        ):
            dt = datetime.datetime.fromtimestamp(float(time.get("data-date")) / 1000.0)
        for span_tz in div.find_all(
            "span", {"class": re.compile(".*match-list-alt__header-timezone.*")}
        ):
            tz_text = span_tz.get_text().strip().replace("(", "").replace(")", "")
            dt = parse(" ".join([str(dt), tz_text]))
        venue_name: str | None = None
        for span in div.find_all(
            "span", {"class": re.compile(".*match-list-alt__header-venue.*")}
        ):
            venue_name = span.get_text().strip().replace(",", "")
        if venue_name is None:
            raise ValueError("venue_name is null")
        if dt is None:
            raise ValueError("dt is null")
        url = None
        for a in div.find_all(
            "a", {"class": re.compile(".*match-list-alt__header-mc-link.*")}
        ):
            url = urllib.parse.urljoin(html_url, a.get("href"))
        yield create_afl_afl_game_model(
            team_names,
            teams_players,
            dt,
            venue_name,
            session,
            ladder,
            url,
            playwright,
        )


def _parse_v2_soup(
    soup: BeautifulSoup, html_url: str
) -> Iterator[
    tuple[list[str], list[list[tuple[str, str, str, str, Position]]], str, str]
]:
    for div in soup.find_all("div", {"class": "team-lineups__item"}):
        team_names = []
        for span in div.find_all(
            "span", {"class": re.compile(".*team-lineups-header__name.*")}
        ):
            team_names = [x.strip() for x in span.get_text().strip().split(" v ")]
        team_players: list[list[tuple[str, str, str, str, Position]]] = [[], []]

        def process_a_player(
            a: Tag,
            idx: int,
            team_players: list[list[tuple[str, str, str, str, Position]]],
            div: Tag,
        ) -> list[list[tuple[str, str, str, str, Position]]]:
            href = a.get("href")
            if not isinstance(href, str):
                raise ValueError("href is not a str")
            player_id = "afl:" + href.split("/")[2]
            name = a.get("title")
            if not isinstance(name, str):
                raise ValueError("name is not a str")
            first_name, sur_name = name.split()
            jersey = None
            for div_shirt in a.find_all(
                "div", {"class": re.compile(".*team-lineups__player-entry--shirt.*")}
            ):
                jersey = div_shirt.get_text().strip()
            if jersey is None:
                raise ValueError("jersey is null")
            position = None
            for span in div.find_all(
                "span", {"class": re.compile(".*team-lineups__position-meta-label.*")}
            ):
                position = position_from_str(span.get_text().strip())
            if position is None:
                raise ValueError("position is null")
            team_players[idx].append(
                (player_id, first_name, sur_name, jersey, position)
            )
            return team_players

        for a in div.find_all(
            "a", {"class": re.compile(".*team-lineups__player-entry--home-team.*")}
        ):
            team_players = process_a_player(a, 0, team_players, div)
        for a in div.find_all(
            "a", {"class": re.compile(".*team-lineups__player-entry--away-team.*")}
        ):
            team_players = process_a_player(a, 1, team_players, div)

        venue_name: str | None = None
        for div_info in div.find_all("div", {"class": "team-lineups-header__info"}):
            header_info = div_info.get_text().strip()
            venue_name = header_info.split("･")[1].strip()
        if venue_name is None:
            raise ValueError("venue_name is null")

        url = None
        for a in div.find_all("a", {"class": "team-lineups-header"}):
            url = urllib.parse.urljoin(html_url, a.get("href"))
        if url is None:
            raise ValueError("url is null")

        yield (team_names, team_players, venue_name, url)


def _parse_v2(
    soup: BeautifulSoup,
    session: requests_cache.CachedSession,
    ladder: list[str],
    html_url: str,
    playwright: Playwright,
) -> Iterator[GameModel]:
    for team_names, team_players, venue_name, url in _parse_v2_soup(soup, html_url):
        yield create_afl_afl_game_model(
            team_names,
            team_players,
            None,
            venue_name,
            session,
            ladder,
            url,
            playwright,
        )


def _parse_game_info(
    html: str,
    session: requests_cache.CachedSession,
    ladder: list[str],
    html_url: str,
    playwright: Playwright,
) -> Iterator[GameModel]:
    soup = BeautifulSoup(html, "lxml")
    found = False
    for game_model in _parse_v1(soup, session, ladder, html_url, playwright):
        found = True
        yield game_model
    if not found:
        yield from _parse_v2(soup, session, ladder, html_url, playwright)


class AFLAFLLeagueModel(LeagueModel):
    """AFL AFL implementation of the league model."""

    def __init__(
        self, session: requests_cache.CachedSession, position: int | None = None
    ) -> None:
        super().__init__(League.AFL, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "afl-afl-league-model"

    @property
    def _ladder(self) -> list[str]:
        ladder = []
        ensure_install()
        with sync_playwright() as p:
            browser = p.chromium.launch()
            context = browser.new_context()
            page = context.new_page()
            try:
                page.goto(
                    "https://www.afl.com.au/ladder",
                    wait_until="networkidle",
                    timeout=60000.0,
                )
            except:  # noqa: E722
                pass
            soup = BeautifulSoup(page.content(), "lxml")
            for span in soup.find_all(
                "span", {"class": re.compile(".*stats-table__club-name.*")}
            ):
                team_name = span.get_text().strip()
                if team_name in ladder:
                    continue
                ladder.append(team_name)
            logging.info("Found ladder: %s", ",".join(ladder))
        return ladder

    @property
    def games(self) -> Iterator[GameModel]:
        ladder = self._ladder
        ensure_install()
        with sync_playwright() as p:
            browser = p.chromium.launch()
            context = browser.new_context()
            page = context.new_page()
            url = "https://www.afl.com.au/matches/team-lineups"
            page.goto(url, wait_until="networkidle")
            yield from _parse_game_info(page.content(), self.session, ladder, url, p)



"""ESPN coach model."""

import datetime

import pytest_is_running
import requests_cache

from ...cache import MEMORY
from ..coach_model import CoachModel


def _create_espn_coach_model(
    session: requests_cache.CachedSession,
    url: str,
) -> CoachModel:
    response = session.get(url)
    response.raise_for_status()
    data = response.json()
    return CoachModel(
        identifier=data["id"],
        name=" ".join([data["firstName"], data["lastName"]]),
        birth_date=None,
        age=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_espn_coach_model(
    session: requests_cache.CachedSession,
    url: str,
) -> CoachModel:
    return _create_espn_coach_model(session, url)


def create_espn_coach_model(
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    url: str,
) -> CoachModel:
    """Create coach model from ESPN."""
    if (
        not pytest_is_running.is_running()
        and dt.date() < datetime.datetime.today().date() - datetime.timedelta(days=7)
    ):
        return _cached_create_espn_coach_model(session, url)
    with session.cache_disabled():
        return _create_espn_coach_model(session, url)



"""NBA Sports DB league model."""

import requests_cache

from ...league import League
from ...sportsdb.sportsdb_league_model import SportsDBLeagueModel


class NBASportsDBLeagueModel(SportsDBLeagueModel):
    """NBA SportsDB implementation of the league model."""

    def __init__(
        self, session: requests_cache.CachedSession, position: int | None = None
    ) -> None:
        super().__init__(session, "4387", League.NBA, position=position)



"""Tests for the social model class."""
import datetime
import unittest

from sportsball.data.social_model import SocialModel


class TestSocialModel(unittest.TestCase):

    def setUp(self):
        self._social_model = SocialModel(
            network="X",
            post="Come see the game",
            comments=10,
            reposts=2,
            likes=100,
            views=None,
            published=datetime.datetime(2010, 1, 1, 10, 10, 0),
        )

    def test_notnull(self):
        self.assertIsNotNone(self._social_model)



"""Sports reference coach model."""

# pylint: disable=duplicate-code
import datetime
import logging
import os
from urllib.parse import urlparse

import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from ...cache import MEMORY
from ...proxy_session import X_NO_WAYBACK
from ..coach_model import CoachModel

_NON_WAYBACK_URLS: set[str] = {
    "https://www.sports-reference.com/cbb/coaches/kelvin-sampson-1.html",
    "https://www.sports-reference.com/cbb/coaches/fred-hoiberg-1.html",
    "https://www.sports-reference.com/cbb/coaches/johnny-dawkins-1.html",
    "https://www.sports-reference.com/cbb/coaches/bruce-pearl-1.html",
    "https://www.sports-reference.com/cbb/coaches/leon-rice-1.html",
    "https://www.sports-reference.com/cbb/coaches/wes-miller-1.html",
    "https://www.sports-reference.com/cbb/coaches/dan-earl-1.html",
}


def _create_sportsreference_coach_model(
    session: requests_cache.CachedSession, coach_url: str, dt: datetime.datetime
) -> CoachModel:
    """Create a coach model from sports reference."""
    headers = {}
    if coach_url in _NON_WAYBACK_URLS:
        headers = {X_NO_WAYBACK: "1"}
    response = session.get(coach_url, headers=headers)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")

    name = None
    for h1 in soup.find_all("h1"):
        for span in h1.find_all("span"):
            name = span.get_text().strip()
            break
        if name is not None:
            break

    if name is None:
        for h1 in soup.find_all("h1"):
            name = h1.get_text().strip()

    if name is None:
        logging.error(response.text)
        logging.error("error url = %s", coach_url)
        raise ValueError("name is null")

    birth_date = None
    for span in soup.find_all("span", {"id": "necro-birth"}):
        birth_date = parse(span.get("data-birth"))

    o = urlparse(coach_url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)
    return CoachModel(
        identifier=identifier,
        name=name,
        birth_date=birth_date,
        age=None if birth_date is None else relativedelta(birth_date, dt).years,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_coach_mode(
    session: requests_cache.CachedSession, coach_url: str, dt: datetime.datetime
) -> CoachModel:
    return _create_sportsreference_coach_model(
        session=session, coach_url=coach_url, dt=dt
    )


def create_sportsreference_coach_model(
    session: requests_cache.CachedSession, coach_url: str, dt: datetime.datetime
) -> CoachModel:
    """Create a coach model from sports reference."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_coach_mode(
            session=session, coach_url=coach_url, dt=dt
        )
    with session.cache_disabled():
        return _create_sportsreference_coach_model(
            session=session, coach_url=coach_url, dt=dt
        )



"""AFL aussportsbetting league model."""

import requests_cache

from ...aussportsbetting.aussportsbetting_league_model import \
    AusSportsBettingLeagueModel
from ...league import League


class AFLAusSportsBettingLeagueModel(AusSportsBettingLeagueModel):
    """AFL AusSportsBetting implementation of the league model."""

    def __init__(
        self, session: requests_cache.CachedSession, position: int | None = None
    ) -> None:
        super().__init__(League.AFL, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "afl-aussportsbetting-league-model"



"""Tests for the HKJC HKJC player model class."""
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.hkjc.hkjc.hkjc_hkjc_player_model import create_hkjc_hkjc_player_model


class TestHKJCHKJCPlayerModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_no_table_for_sire(self):
        with requests_mock.Mocker() as m:
            url = "https://racing.hkjc.com/racing/information/english/Horse/SameSire.aspx?HorseSire=Nyquist"
            with open(os.path.join(self.dir, "sire_nyquist_2.html"), "rb") as f:
                m.get(url, content=f.read())
            player_model = create_hkjc_hkjc_player_model(
                session=self._session,
                url=url,
                jersey=None,
                handicap_weight=None,
                starting_position=None,
                weight=None,
            )
            self.assertEqual(player_model.name, "Nyquist")



"""The prototype class for a venue."""

from typing import Literal

from pydantic import BaseModel, Field

from .address_model import AddressModel
from .field_type import TYPE_KEY, FieldType

VENUE_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
VENUE_ADDRESS_COLUMN: Literal["address"] = "address"
VENUE_IS_TURF_COLUMN: Literal["is_turf"] = "is_turf"
VENUE_IS_DIRT_COLUMN: Literal["is_dirt"] = "is_dirt"


class VenueModel(BaseModel):
    """The serialisable venue class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=VENUE_IDENTIFIER_COLUMN,
    )
    name: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.TEXT})
    address: AddressModel | None = Field(..., alias=VENUE_ADDRESS_COLUMN)
    is_grass: bool | None
    is_indoor: bool | None
    is_turf: bool | None = Field(..., alias=VENUE_IS_TURF_COLUMN)
    is_dirt: bool | None = Field(..., alias=VENUE_IS_DIRT_COLUMN)
